from .finding import _Finding, _CoreFinding
from .wp_item_version import WPItemVersion


class Timthumb(_Finding, _CoreFinding):
    def __init__(self, url, data, *args, **kwargs):
        """From https://github.com/wpscanteam/wpscan/blob/master/app/views/json/enumeration/timthumbs.erb"""

        super().__init__(data, *args, **kwargs)
        self.url = url
        self.version = WPItemVersion(self.data.get("version", None), *args, **kwargs)

    def get_infos(self):
        """Return 1 info"""
        info = "Timthumb: {}".format(self.url)
        # If finding infos are present and show_all_details, add them
        if super().get_infos()[0] and self.show_all_details:
            info += "\n{}".format(super().get_infos()[0])
        if self.version.get_infos():
            info += "\n{}".format(self.version.get_infos()[0])
        else:
            info += "\nThe version could not be determined"
        return [info]

    def get_warnings(self):
        """Return empty list"""
        return []

    def get_alerts(self):
        """Return timthumb vulnerabilities"""
        return [
            "Timthumb {}".format(alert)
            for alert in super().get_alerts() + self.version.get_alerts()
        ]

    def get_version(self):
        return self.version.get_version()

    def get_version_status(self):
        if self.version.get_infos():
            return "Unknown"
        else:
            return "N/A"

    def get_vulnerabilities_string(self):
        return "{}{}".format(
            len(self.vulnerabilities),
            " (potential)"
            if not self.version.get_infos() and super().get_alerts()
            else "",
        )

    def get_name(self):
        return "Timthumb"



from .finding import _Finding, _CoreFinding


class WordPressVersion(_Finding, _CoreFinding):
    def __init__(self, data, *args, **kwargs):
        """From https://github.com/wpscanteam/wpscan/blob/master/app/views/json/wp_version/version.erb"""

        super().__init__(data, *args, **kwargs)
        self.number = self.data.get("number", None)
        self.release_date = self.data.get("release_date", None)
        self.status = self.data.get("status", None)

    def _get_infos(self):
        """Return 1 info"""
        if self.number:
            info = "Wordpress version: {}".format(self.number)
            if self.status == "latest":
                info += " (up to date)"
            if self.release_date:
                info += "\nRelease Date: {}".format(self.release_date)
            # Not showing "Status: "
            # if self.status:
            #     info+="\nStatus: {}".format(self.status.title())
        else:
            info = "The WordPress version could not be detected"
        # If finding infos are present and show_all_details, add them
        if super().get_infos()[0] and self.show_all_details:
            info += "\n{}".format(super().get_infos()[0])
        return [info]

    def get_infos(self):
        """Return 0 or 1 info, no infos if WordPressVersion triggedred warning, use get_warnings()"""
        if not self.get_warnings():
            return self._get_infos()
        else:
            return []

    def get_warnings(self):
        """Return 0 or 1 warning"""

        if self.status == "insecure":
            warning = "Outdated "
            warning += self._get_infos()[0]
            return [warning]
        else:
            return []

    def get_alerts(self):
        """Return Wordpress Version vulnerabilities"""
        return ["Wordpress {}".format(alert) for alert in super().get_alerts()]

    def get_version(self):
        """Return the version string or 'Unknown'"""
        return self.number if self.number else "Unknown"

    def get_version_status(self):
        if self.number:
            if self.status == "insecure":
                return "Outdated"
            elif self.status == "latest":
                return "Latest"
            else:
                return "Unknown"
        else:
            return "N/A"

    def get_vulnerabilities_string(self):
        return "{}".format(len(self.vulnerabilities))

    def get_name(self):
        return "WordPress {} {}".format(
            self.get_version(),
            "({})".format(self.release_date)
            if self.release_date and self.number
            else "",
        )



from abc import ABC, abstractmethod

from ..base import _Component


class _Finding(_Component):
    """ Generic WPScan finding"""

    def __init__(self, data, *args, **kwargs):
        """From https://github.com/wpscanteam/wpscan/blob/master/app/views/json/finding.erb"""

        super().__init__(data, *args, **kwargs)

        self.found_by = self.data.get("found_by", None)
        self.confidence = self.data.get("confidence", None)
        self.interesting_entries = self.data.get("interesting_entries", None)
        self.confirmed_by = self.data.get("confirmed_by", None)
        self.vulnerabilities = [
            Vulnerability(vuln, *args, **kwargs)
            for vuln in self.data.get("vulnerabilities", [])
        ]
        self.references = self.data.get("references", None)

    def get_alerts(self):
        """Return list of vulnerabilities"""
        alerts = []
        for v in self.vulnerabilities:
            alerts.extend(v.get_alerts())
        return alerts

    def get_infos(self):
        """Return 1 info, only interesting entries. If no interesting entries: return an empty info string (to avoid errors)"""
        info = ""
        if self.interesting_entries:
            info += "Interesting entries: \n- {}".format(
                "\n- ".join(self.interesting_entries)
            )
            if self.show_all_details:
                info += "\n"
        if self.show_all_details:
            if self.found_by:
                info += "Found by: {} ".format(self.found_by)
            if self.confidence:
                info += "(confidence: {})".format(self.confidence)
            if self.confirmed_by:
                info += "\nConfirmed by: "
                for entry in self.confirmed_by:
                    info += "\n- {} ".format(entry)
                    if self.confirmed_by[entry].get("confidence", None):
                        info += "(confidence: {})".format(
                            self.confirmed_by[entry]["confidence"]
                        )
                    if self.confirmed_by.get("interesting_entries", None):
                        info += "\n  Interesting entries: \n  - {}".format(
                            "\n  - ".join(self.confirmed_by.get("interesting_entries"))
                        )

        return [info]

    def get_references_str(self):
        """Process CVE, WPVulnDB, ExploitDB and Metasploit references to add links"""
        alert = ""
        if self.references:
            alert += "References: "
            for ref in self.references:
                if ref == "cve":
                    for cve in self.references[ref]:
                        alert += "\n- CVE: https://cve.mitre.org/cgi-bin/cvename.cgi?name=CVE-{}".format(
                            cve
                        )
                elif ref == "wpvulndb":
                    for wpvulndb in self.references[ref]:
                        alert += "\n- WPVulnDB: https://wpvulndb.com/vulnerabilities/{}".format(
                            wpvulndb
                        )
                elif ref == "metasploit":
                    for metasploit in self.references[ref]:
                        alert += "\n- Metasploit: https://www.rapid7.com/db/modules/{}".format(
                            metasploit
                        )
                elif ref == "exploitdb":
                    for exploitdb in self.references[ref]:
                        alert += "\n- ExploitDB: https://www.exploit-db.com/exploits/{}".format(
                            exploitdb
                        )
                elif ref == "packetstorm":
                    for packetstorm in self.references[ref]:
                        alert += "\n- Packetstorm: https://packetstormsecurity.com/files/{}".format(
                            packetstorm
                        )

                else:
                    for link in self.references[ref]:
                        alert += "\n- {}: {}".format(ref.title(), link)
        return alert


class _CoreFinding(ABC):
    @abstractmethod
    def get_version(self):
        """Return the version number (as string) or "Unknown" or "" if not applicable. """
        pass

    @abstractmethod
    def get_version_status(self):
        """Return a string in : "Outdated", "Latest", "Unknown" or "" if not applicable.  """
        pass

    @abstractmethod
    def get_vulnerabilities_string(self):
        """Return the number of vulnerabilities (as string) with indications if need be. """
        pass

    @abstractmethod
    def get_name(self):
        """Return the name of the finding. """
        pass

    def get_status(self):
        """Return a string in : "Alert", "Warning", "Ok", "Unknown" """
        try:
            return (
                "Alert"
                if self.get_alerts()
                else "Warning"
                if self.get_warnings()
                else "Ok"
                if self.get_version_status() == "Latest"
                else "Unknown"
            )
        except AttributeError:
            return None


class _CoreFindingNoVersion(_CoreFinding):
    def get_version(self):
        """ Returns empty string"""
        return ""

    def get_version_status(self):
        """ Returns empty string"""
        return ""

    def get_vulnerabilities_string(self):
        """ Returns empty string"""
        return ""


# Class Vulnerability moved with Finding to avoid circular imorts errors
class Vulnerability(_Finding):
    def __init__(self, data, *args, **kwargs):
        """From https://github.com/wpscanteam/wpscan/blob/master/app/views/json/finding.erb"""

        super().__init__(data, *args, **kwargs)

        self.title = self.data.get("title", None)
        self.cvss = self.data.get("cvss", None)
        self.fixed_in = self.data.get("fixed_in", None)

    def get_alerts(self):
        """Return 1 alert. First line of alert string contain the vulnerability title. Process CVE, WPVulnDB, ExploitDB and Metasploit references to add links"""
        alert = "Vulnerability: {}".format(self.title)

        if self.cvss:
            alert += "\nCVSS: {}".format(self.cvss)
        if self.fixed_in:
            alert += "\nFixed in: {}".format(self.fixed_in)
        else:
            alert += "\nNo known fix"
        if self.references:
            alert += "\n{}".format(self.get_references_str())
        return [alert]

    def get_warnings(self):
        """Return empty list"""
        return []

    def get_infos(self):
        """Return empty list"""
        return []



from .wp_item import WPItem


class Plugin(WPItem):
    def __init__(self, data, *args, **kwargs):
        """From https://github.com/wpscanteam/wpscan/blob/master/app/views/json/enumeration/plugins.erb"""

        super().__init__(data, *args, **kwargs)

    def get_infos(self):
        """Return 1 or 0 info if pluging trigerred warning"""
        return ["Plugin: {}".format(info) for info in super().get_infos()]

    def get_warnings(self):
        """Return plugin warnings"""
        # Adds plugin prefix on all warnings except vulns
        return [
            "{}{}".format(
                "Plugin: " if "Vulnerability" not in warning.splitlines()[0] else "",
                warning,
            )
            for warning in super().get_warnings()
        ]

    def get_name(self):
        return "Plugin: {}".format(self.slug)



import re

from .base import _Parser
from .components import InterestingFinding
from .results import _WPScanResults

####################  CLI PARSER ######################


class WPScanCliParser(_Parser):
    """Main interface to parse WPScan CLI output.

    - wpscan_output: WPScan output as string.
    - false_positives_strings: List of false positive strings.

    """

    def __init__(self, wpscan_output, false_positives_strings=None):

        if not wpscan_output:
            wpscan_output = ""
        # _Parser config: false positives string and verbosity (not available with cli parser)
        parser_config = dict(
            false_positives_strings=false_positives_strings, show_all_details=False
        )
        super().__init__(wpscan_output, **parser_config)
        self.infos, self.warnings, self.alerts = self.parse_cli(wpscan_output)

    def get_infos(self):
        """ Return all the parsed infos"""
        return self.infos

    def get_warnings(self):
        """ Return all the parsed warnings"""
        return self.warnings

    def get_alerts(self):
        """ Return all the parsed alerts"""
        return self.alerts

    def _parse_cli_toogle(self, line, warning_on, alert_on):
        # Color parsing
        if "33m[!]" in line:
            warning_on = True
        elif "31m[!]" in line:
            alert_on = True
        # No color parsing Warnings string are hard coded here
        elif "[!]" in line and any(
            [
                m in line
                for m in [
                    "The version is out of date",
                    "No WPVulnDB API Token given",
                    "You can get a free API token",
                ]
            ]
        ):
            warning_on = True
        elif "[!]" in line:
            alert_on = True
        # Both method with color and no color apply supplementary proccessing
        # Warning for insecure Wordpress and based on interesting findings strings
        if any(
            string in line
            for string in ["Insecure"]
            + InterestingFinding.INTERESTING_FINDING_WARNING_STRINGS
        ):
            warning_on = True
        # Trigger alert based on interesting finding alert strings
        if any(
            string in line
            for string in InterestingFinding.INTERESTING_FINDING_ALERT_STRINGS
        ):
            alert_on = True
        # Lower voice of Vulnerabilities found but not plugin version
        if "The version could not be determined" in line and alert_on:
            alert_on = False
            warning_on = True
        return (warning_on, alert_on)

    def _ignore_false_positives(self, infos, warnings, alerts):
        """Process false positives"""
        for alert in warnings + alerts:
            if self.is_false_positive(alert):
                try:
                    alerts.remove(alert)
                except ValueError:
                    warnings.remove(alert)
                infos.append("[False positive]\n{}".format(alert))

        return infos, warnings, alerts

    def parse_cli(self, wpscan_output):
        """Parse the ( messages, warnings, alerts ) from WPScan CLI output string.
        Return results as tuple( messages, warnings, alerts )."""
        # Init scan messages
        (messages, warnings, alerts) = ([], [], [])
        # Init messages toogles
        warning_on, alert_on = False, False
        message_lines = []
        current_message = ""

        # Every blank ("") line will be considered as a message separator
        for line in wpscan_output.splitlines() + [""]:

            # Parse all output lines and build infos, warnings and alerts
            line = line.strip()

            # Parse line
            warning_on, alert_on = self._parse_cli_toogle(line, warning_on, alert_on)

            # Remove colorization anyway after parsing
            line = re.sub(r"(\x1b|\[[0-9][0-9]?m)", "", line)
            # Append line to message. Handle the begin of the message case
            message_lines.append(line)

            # Build message
            current_message = "\n".join(
                [m for m in message_lines if m not in ["", "|"]]
            ).strip()

            # Message separator just a white line.
            # Only if the message if not empty.
            if line.strip() not in [""] or current_message.strip() == "":
                continue

            # End of the message

            # Post process message to separate ALERTS into different messages of same status and add rest of the infos to warnings
            if (alert_on or warning_on) and any(
                s in current_message
                for s in ["vulnerabilities identified", "vulnerability identified"]
            ):
                messages_separated = []
                msg = []
                for l in message_lines + ["|"]:
                    if l.strip() == "|":
                        messages_separated.append(
                            "\n".join([m for m in msg if m not in ["", "|"]])
                        )
                        msg = []
                    msg.append(l)

                # Append Vulnerabilities messages to ALERTS and other infos in one message
                vulnerabilities = [
                    m for m in messages_separated if "| [!] Title" in m.splitlines()[0]
                ]

                # Add the plugin infos to warnings or false positive if every vulnerabilities are ignore
                plugin_infos = "\n".join(
                    [
                        m
                        for m in messages_separated
                        if "| [!] Title" not in m.splitlines()[0]
                    ]
                )

                if (
                    len([v for v in vulnerabilities if not self.is_false_positive(v)])
                    > 0
                    and "The version could not be determined" in plugin_infos
                ):
                    warnings.append(
                        plugin_infos + "\nAll known vulnerabilities are listed"
                    )
                else:
                    messages.append(plugin_infos)

                if alert_on:
                    alerts.extend(vulnerabilities)
                elif warning_on:
                    warnings.extend(vulnerabilities)

            elif warning_on:
                warnings.append(current_message)
            else:
                messages.append(current_message)
            message_lines = []
            current_message = ""
            # Reset Toogle Warning/Alert
            warning_on, alert_on = False, False

        return self._ignore_false_positives(messages, warnings, alerts)

    def get_error(self):
        if "Scan Aborted" in self.data:
            return "WPScan failed: {}".format(
                "\n".join(
                    line for line in self.data.splitlines() if "Scan Aborted" in line
                )
            )
        else:
            return None

    def get_results(self):
        """
        Returns a dictionnary structure like:
        ```
        {
        'infos':[],
        'warnings':[],
        'alerts':[],
        'summary':{
            'table':None,
            'line':'WPScan result summary: alerts={}, warnings={}, infos={}, error={}'
            },
        'error':None
        }
        ```
        """
        results = _WPScanResults()
        results["infos"] = self.get_infos()
        results["warnings"] = self.get_warnings()
        results["alerts"] = self.get_alerts()
        results["summary"]["line"] = self.get_summary_line()
        results["error"] = self.get_error()
        return dict(results)



from ..base import _Component


class ScanFinished(_Component):
    def __init__(self, data, *args, **kwargs):
        """From https://github.com/wpscanteam/CMSScanner/blob/master/app/views/json/core/finished.erb"""

        super().__init__(data, *args, **kwargs)

        self.stop_time = self.data.get("stop_time", None)
        self.elapsed = self.data.get("elapsed", None)
        self.requests_done = self.data.get("requests_done", None)
        self.cached_requests = self.data.get("cached_requests", None)
        self.data_sent_humanised = self.data.get("data_sent_humanised", None)
        self.data_received_humanised = self.data.get("data_received_humanised", None)
        self.used_memory_humanised = self.data.get("used_memory_humanised", None)

    def get_infos(self):
        """Return 1 Scan Finished info"""

        info = "Scan duration: {} seconds".format(self.elapsed)
        if self.show_all_details:
            info += "\nStop Time: {}".format(self.stop_time)
            info += "\nRequests Done: {}".format(self.requests_done)
            info += "\nCached Requests: {}".format(self.cached_requests)
            info += "\nData Sent: {}".format(self.data_sent_humanised)
            info += "\nData Received: {}".format(self.data_received_humanised)
            info += "\nUsed Memory: {}".format(self.used_memory_humanised)

        return [info]

    def get_warnings(self):
        """Return empty list"""
        return []

    def get_alerts(self):
        """Return empty list"""
        return []



import unittest
from wpscan_out_parse import parse_results_from_string

class T(unittest.TestCase):
    
       def test_parser(self):
              # false positives
              out = open("test/output_files/wordpress_no_vuln.json").read()
              result=parse_results_from_string(out)
              self.assertEqual(0, len(result['alerts']))

              out = open("test/output_files/wordpress_one_vuln.json").read()
              result=parse_results_from_string(out)
              self.assertEqual(4, len(result['warnings']))

              out = open("test/output_files/wordpress_many_vuln.json").read()
              result=parse_results_from_string(out)
              self.assertEqual(1, len(result['alerts']))

              out = open("test/output_files/wordpress_no_vuln.txt").read()
              result=parse_results_from_string(out)
              self.assertEqual(0, len(result['alerts']))

              out = open("test/output_files/wordpress_one_warning.txt").read()
              result=parse_results_from_string(out)
              self.assertEqual(2, len(result['warnings']))

              out = open("test/output_files/wordpress_many_vuln.txt").read()
              result=parse_results_from_string(out)
              self.assertEqual(8, len(result['alerts']))
              
              out = open("test/output_files/wordpress_one_vuln.txt").read()
              result=parse_results_from_string(out)
              self.assertEqual(1, len(result['alerts']))

    #    def test_false_positives(self):
    #           pass

    #    def test_version_could_not_be_detected_warning(self): 
    #           pass

    #    def test_version_could_not_be_detected_false_positive_mgmt(self): 
    #           pass

    #    def test_oudated_plugin_or_theme_version_warning(self):
    #           pass

    #    def test_vulnerabilities(self):
    #           pass

    #    def test_insecure_wordpress_warning(self):
    #           pass

    #    def test_password_attack(self):
    #           pass

    #    def test_lots_of_enumeration(self):
    #           pass

    #    def test_ref_metasploit_cve_exploitdb(self):
    #           pass


from .finding import _Finding


class InterestingFinding(_Finding):

    INTERESTING_FINDING_WARNING_STRINGS = [
        "Upload directory has listing enabled",
        "Registration is enabled",
        "Debug Log found",
        "codex.wordpress.org/Debugging_in_WordPress",
        "Fantastico list found",
        "www.acunetix.com/vulnerabilities/fantastico-fileslist/",
    ]

    INTERESTING_FINDING_ALERT_STRINGS = [
        "SQL Dump found",
        "Full Path Disclosure found",
        "www.owasp.org/index.php/Full_Path_Disclosure",
        "codex.wordpress.org/Resetting_Your_Password#Using_the_Emergency_Password_Reset_Script",
        "www.exploit-db.com/ghdb/3981/",
        "A backup directory has been found",
        "github.com/wpscanteam/wpscan/issues/422",
        "ThemeMakers migration file found",
        "packetstormsecurity.com/files/131957",
        "Search Replace DB script found",
        "interconnectit.com/products/search-and-replace-for-wordpress-databases/",
    ]

    def __init__(self, data, *args, **kwargs):
        """From https://github.com/wpscanteam/CMSScanner/blob/master/app/views/json/interesting_findings/findings.erb
        Warnings and Alerts strings are from https://github.com/wpscanteam/wpscan/blob/master/app/models/interesting_finding.rb
        """

        super().__init__(data, *args, **kwargs)
        self.url = self.data.get("url", None)
        self.to_s = self.data.get("to_s", None)
        self.type = self.data.get("type", None)

    def _get_infos(self):
        """Return 1 info. First line of info string is the to_s string or the finding type. Complete references links too."""
        info = ""
        if self.to_s != self.url:
            info += self.to_s
        elif self.type:
            info += self.type.title()
        if self.url and self.url not in self.to_s:
            info += "\nURL: {}".format(self.url)
        # If finding infos are present, add them
        if super().get_infos()[0]:
            info += "\n{}".format(super().get_infos()[0])
        if self.references:
            info += "\n{}".format(self.get_references_str())
        return [info]

    def get_infos(self):
        """Return 1 info or 0 if finding is a warning or an alert"""
        return [
            info
            for info in self._get_infos()
            if not any(
                [
                    string in info
                    for string in self.INTERESTING_FINDING_WARNING_STRINGS
                    + self.INTERESTING_FINDING_ALERT_STRINGS
                ]
            )
        ]

    def get_warnings(self):
        """Return list of warnings if finding match warning string"""
        return [
            info
            for info in self._get_infos()
            if any(
                [string in info for string in self.INTERESTING_FINDING_WARNING_STRINGS]
            )
        ]

    def get_alerts(self):
        """Return list of alerts if finding match ALERT string"""
        return [
            info
            for info in self._get_infos()
            if any(
                [string in info for string in self.INTERESTING_FINDING_ALERT_STRINGS]
            )
        ]



from .finding import _Finding, _CoreFinding
from .wp_item_version import WPItemVersion


class WPItem(_Finding, _CoreFinding):
    def __init__(self, data, *args, **kwargs):
        """From https://github.com/wpscanteam/wpscan/blob/master/app/views/json/wp_item.erb"""

        super().__init__(data, *args, **kwargs)

        self.slug = self.data.get("slug", None)
        self.location = self.data.get("location", None)
        self.latest_version = self.data.get("latest_version", None)
        self.last_updated = self.data.get("last_updated", None)
        self.outdated = self.data.get("outdated", None)
        self.readme_url = self.data.get("readme_url", None)
        self.directory_listing = self.data.get("directory_listing", None)
        self.error_log_url = self.data.get("error_log_url", None)
        self.version = WPItemVersion(self.data.get("version", None), *args, **kwargs)

    def _get_warnings(self):
        """Return 0 or 1 warning. The warning can contain infos about oudated plugin, directory listing or accessible error log.
        First line of warning string is the plugin slug. Location also added as a reference."""
        # Test if there is issues
        issue_data = ""
        if self.outdated:
            issue_data += "\nThe version is out of date"
        if self.directory_listing:
            issue_data += "\nDirectory listing is enabled"
            issue_data += "\nLocation: {}".format(self.location)
        if self.error_log_url:
            issue_data += "\nAn error log file has been found: {}".format(
                self.error_log_url
            )

        if not issue_data:
            return []  # Return if no issues
        else:
            return [issue_data]

    def get_alerts(self):
        """Return list of know plugin or theme vulnerability. Empty list is returned if plugin version is unrecognized"""
        alerts = []
        if self.version.get_infos():
            alerts.extend(["{}".format(alert) for alert in super().get_alerts()])
            alerts.extend(["{}".format(alert) for alert in self.version.get_alerts()])
        return alerts

    def get_warnings(self):
        """Return plugin or theme warnings, if oudated plugin, directory listing, accessible error log and
        for all know vulnerabilities if plugin version could not be recognized.
        Adds a special text saying the version is unrecognized if that's the case"""
        warnings = []
        # Prepare warning string
        warning = self.slug if self.slug else str()
        # Get oudated theme warning
        if self._get_warnings():
            warning += self._get_warnings()[0]
        if warning:
            warning += "\n"
        # Get generic infos
        warning += self._get_infos()[0]
        # If vulns are found and the version is unrecognized
        if not self.version.get_infos() and super().get_alerts():
            # Adds a special text saying all vulns are listed
            warning += "\nAll known vulnerabilities are listed"
        # If vulns are found and the version is unrecognized or other issue like outdated version or directory listing enable
        if (
            not self.version.get_infos() and super().get_alerts()
        ) or self._get_warnings():
            warnings.append(warning)
        # If vulns are found and the version is unrecognized : add Potential vulns
        if not self.version.get_infos() and super().get_alerts():
            warnings.extend(
                ["Potential {}".format(warn) for warn in super().get_alerts()]
            )
            warnings.extend(
                ["Potential {}".format(warn) for warn in self.version.get_alerts()]
            )
        return warnings

    def _get_infos(self):
        """Return 1 info"""
        info = ""
        if self.show_all_details:
            if self.location:
                info += "Location: {}".format(self.location)
            if self.last_updated:
                info += "\nLast Updated: {}".format(self.last_updated)
        if self.readme_url:
            if info:
                info += "\n"
            info += "Readme: {}".format(self.readme_url)
        # If finding infos are present, add them
        if info:
            info += "\n"
        if super().get_infos()[0] and self.show_all_details:
            info += "{}\n".format(super().get_infos()[0])
        if self.version.get_infos():
            info += self.version.get_infos()[0]
            if self.version.number == self.latest_version:
                info += " (up to date)"
            elif self.latest_version:
                info += " (latest is {})".format(self.latest_version)
        else:
            info += "The version could not be determined"
            if self.latest_version:
                info += " (latest is {})".format(self.latest_version)

        return [info]

    def get_infos(self):
        """Return 0 or 1 info, no info if WPItem triggered warning, use get_warnings()"""
        if not self.get_warnings():
            return ["{}\n{}".format(self.slug, self._get_infos()[0])]
        else:
            return []

    def get_version(self):
        return self.version.get_version()

    def get_version_status(self):
        if self.version.get_infos():
            if self.outdated:
                return "Outdated"
            elif self.version.number == self.latest_version:
                return "Latest"
            else:
                return "Unknown"
        else:
            return "N/A"

    def get_vulnerabilities_string(self):
        return "{}{}".format(
            len(self.vulnerabilities),
            " (potential)"
            if not self.version.get_infos() and super().get_alerts()
            else "",
        )



from .finding import _Finding


class User(_Finding):
    def __init__(self, username, data, *args, **kwargs):
        """From https://github.com/wpscanteam/wpscan/blob/master/app/views/json/enumeration/users.erb
        And https://github.com/wpscanteam/wpscan/blob/master/app/views/json/password_attack/users.erb
        """

        super().__init__(data, *args, **kwargs)

        self.username = username
        self.id = self.data.get("id", None)
        self.password = self.data.get("password", None)

    def get_infos(self):
        """Return 1 info"""
        info = "User Identified: {}".format(self.username)
        if self.id:
            info += "\nID: {}".format(self.id)
        # If finding infos are present, add them
        if super().get_infos()[0] and self.show_all_details:
            info += "\n{}".format(super().get_infos()[0])
        return [info]

    def get_warnings(self):
        """Return empty list"""
        return []

    def get_alerts(self):
        """Return 0 or 1 alert. Alert if password found. Used by PasswordAttack component"""
        if self.password:
            alert = "Username: {}".format(self.username)
            alert += "Password: {}".format(self.password)
            return [alert]
        else:
            return []



from .finding import _Finding, _CoreFindingNoVersion


class DBExport(_Finding, _CoreFindingNoVersion):
    def __init__(self, url, data, *args, **kwargs):
        """From https://github.com/wpscanteam/wpscan/blob/master/app/views/json/enumeration/db_exports.erb"""

        super().__init__(data, *args, **kwargs)
        self.url = url

    def get_alerts(self):
        """Return 1 DBExport alert"""
        alert = "Database Export: {}".format(self.url)
        # If finding infos are present, add them
        if super().get_infos()[0] and self.show_all_details:
            alert += "\n{}".format(super().get_infos()[0])
        return [alert]

    def get_warnings(self):
        """Return empty list"""
        return []

    def get_infos(self):
        """Return empty list"""
        return []

    def get_name(self):
        return "Database Export"



from .theme import Theme


class MainTheme(Theme):
    def __init__(self, data, *args, **kwargs):
        """From https://github.com/wpscanteam/wpscan/blob/master/app/views/json/main_theme/theme.erb"""

        super().__init__(data, *args, **kwargs)

    def get_infos(self):
        """Return 1 info"""
        return [
            "Main Theme: {}".format(info) for info in super(Theme, self).get_infos()
        ]

    def get_warnings(self):
        """Return Main Theme warnings"""
        return [
            "{}{}".format(
                "Main Theme: "
                if "Vulnerability" not in warning.splitlines()[0]
                else "",
                warning,
            )
            for warning in super(Theme, self).get_warnings()
        ]

    def get_name(self):
        return "Main Theme: {}".format(self.slug)



import json
import re


def format_results(results, format):
    """
    Format the results dict into a "html", "cli" or "json" string.

    - results: resutlts dict objject.
    - format: in `"html"`, `"cli"` or `"json"`
    """
    if format == "json":
        return json.dumps(dict(results), indent=4)
    else:
        return build_message(dict(results), format=format)


def build_message(results, warnings=True, infos=True, format="cli"):
    """Build mail message text base on report and warnngs and info switch"""
    message = ""
    if results["error"]:
        message += "An error occurred.\n".replace(
            "\n", "<br/>\n" if format == "html" else "\n"
        )
    elif results["alerts"]:
        message += "Vulnerabilities have been detected by WPScan.\n".replace(
            "\n", "<br/>\n" if format == "html" else "\n"
        )
    elif results["warnings"]:
        message += "Issues have been detected by WPScan.\n".replace(
            "\n", "<br/>\n" if format == "html" else "\n"
        )

    if results["summary"] and results["summary"]["line"]:
        summary = []

        if format == "cli":
            if results["summary"]["table"]:
                summary.append(
                    format_summary_ascii_table(
                        results["summary"]["table"], results["summary"]["line"]
                    )
                )
            else:
                summary.append(results["summary"]["line"])
        elif format == "html":
            if results["summary"]["table"]:
                summary.append(
                    format_summary_html(
                        results["summary"]["table"], results["summary"]["line"]
                    )
                )
            else:
                summary.append(results["summary"]["line"])
        else:
            raise ValueError("format can only be 'cli' or 'html'")

        message += format_issues(
            "Summary", summary, format=format, apply_br_tab_replace_on_issues=False
        )

    if results["error"]:
        message += "\n".replace("\n", "<br/>\n" if format == "html" else "\n")
        message += format_issues("Error", [results["error"]], format=format)

    if results["alerts"]:
        message += "\n".replace("\n", "<br/>\n" if format == "html" else "\n")
        message += format_issues("Alerts", results["alerts"], format=format)

    if results["warnings"] and warnings:
        message += "\n".replace("\n", "<br/>\n" if format == "html" else "\n")
        message += format_issues("Warnings", results["warnings"], format=format)

    if results["infos"] and infos:
        message += "\n".replace("\n", "<br/>\n" if format == "html" else "\n")
        message += format_issues("Informations", results["infos"], format=format)

    if format == "html":
        message = "<div>" + message + "</div>"

    return message


def format_issues(title, issues, format="cli", apply_br_tab_replace_on_issues=True):
    """Format one block of issues to text with the title"""
    message = ""
    if issues:
        if format == "html":
            message += (
                "<br/>\n&nbsp;&nbsp;&nbsp;&nbsp;%s<br/>\n&nbsp;&nbsp;&nbsp;&nbsp;%s<br/>\n<br/>\n"
                % (title, "-" * len(title))
            )
            message += "<br/>\n<br/>\n".join(
                [
                    replace(issue, {"\n": "<br/>\n", "\t": "&nbsp;&nbsp;&nbsp;&nbsp;"})
                    if format == "html" and apply_br_tab_replace_on_issues
                    else issue
                    for issue in issues
                ]
            )
        else:
            message += "\n\t%s\n\t%s\n\n" % (title, "-" * len(title)) + "\n\n".join(
                issues
            )
    return message


def format_summary_ascii_table(table, line):
    """Return a nice string table
    Author: Thierry Husson - Use it as you want but don't blame me.
    """
    myDict = table
    colList = ["Component", "Version", "Version State", "Vulnerabilities", "Status"]
    myList = [colList]  # 1st row = header
    for item in myDict:
        myList.append(
            [str(item[col] if item[col] is not None else "") for col in colList]
        )
    colSize = [max(map(len, col)) for col in zip(*myList)]
    formatStr = " | ".join(["{{:<{}}}".format(i) for i in colSize])
    myList.insert(1, ["-" * i for i in colSize])  # Seperating line
    string = "\n".join(formatStr.format(*item) for item in myList)
    return string + "\n\n" + line


def format_summary_html(table, line):
    row_fmt = """      <tr>
        <td>{}</td>
        <td>{}</td>
        <td>{}</td>
        <td>{}</td>
        <td>{}</td>
    </tr>
    """
    table_rows = ""
    for row in table:
        table_rows += row_fmt.format(
            row["Component"],
            row["Version"],
            row["Version State"],
            row["Vulnerabilities"],
            # Determine status color
            '<b style="color:{color}">{status}</b>'.format(
                status=row["Status"],
                color="#8B0000"
                if row["Status"] == "Alert"
                else "#FFD700"
                if row["Status"] == "Warning"
                else "#228B22"
                if row["Status"] == "Ok"
                else "#996633"
                if row["Status"] == "Unknown"
                else "#000000",
            ),
        )

    return """<table>
        <tr>
            <th>Component</th>
            <th>Version</th>
            <th>Version State</th>
            <th>Vulnerabilities</th>
            <th>Status</th>
        </tr>
    {}
    </table>
    <br/>{}""".format(
        table_rows, line
    )


def replace(text, conditions):
    """Multiple replacements helper method.  Stolen on the web"""
    rep = conditions
    rep = dict((re.escape(k), rep[k]) for k in rep)
    pattern = re.compile("|".join(rep.keys()))
    text = pattern.sub(lambda m: rep[re.escape(m.group(0))], text)
    return text
